from manim_ml.neural_network.activation_functions.relu import ReLUFunction
from manim_ml.neural_network.activation_functions.sigmoid import SigmoidFunction

name_to_activation_function_map = {
    "ReLU": ReLUFunction,
    "Sigmoid": SigmoidFunction
}

def get_activation_function_by_name(name):
    assert name in name_to_activation_function_map.keys(), \
        f"Unrecognized activation function {name}"

    return name_to_activation_function_map[name]



from manim import *
from manim_ml.neural_network.layers import NeuralNetworkLayer
from manim_ml.image import GrayscaleImageMobject, LabeledColorImage
import numpy as np


class TripletLayer(NeuralNetworkLayer):
    """Shows triplet images"""

    def __init__(
        self,
        anchor,
        positive,
        negative,
        stroke_width=5,
        font_size=22,
        buff=0.2,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.anchor = anchor
        self.positive = positive
        self.negative = negative
        self.buff = buff

        self.stroke_width = stroke_width
        self.font_size = font_size

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs
    ):
        # Make the assets
        self.assets = self.make_assets()
        self.add(self.assets)

    @classmethod
    def from_paths(
        cls,
        anchor_path,
        positive_path,
        negative_path,
        grayscale=True,
        font_size=22,
        buff=0.2,
    ):
        """Creates a triplet using the anchor paths"""
        # Load images from path
        if grayscale:
            anchor = GrayscaleImageMobject.from_path(anchor_path)
            positive = GrayscaleImageMobject.from_path(positive_path)
            negative = GrayscaleImageMobject.from_path(negative_path)
        else:
            anchor = ImageMobject(anchor_path)
            positive = ImageMobject(positive_path)
            negative = ImageMobject(negative_path)
        # Make the layer
        triplet_layer = cls(anchor, positive, negative, font_size=font_size, buff=buff)

        return triplet_layer

    def make_assets(self):
        """
        Constructs the assets needed for a triplet layer
        """
        # Handle anchor
        anchor_group = LabeledColorImage(
            self.anchor,
            color=WHITE,
            label="Anchor",
            stroke_width=self.stroke_width,
            font_size=self.font_size,
            buff=self.buff,
        )
        # Handle positive
        positive_group = LabeledColorImage(
            self.positive,
            color=GREEN,
            label="Positive",
            stroke_width=self.stroke_width,
            font_size=self.font_size,
            buff=self.buff,
        )
        # Handle negative
        negative_group = LabeledColorImage(
            self.negative,
            color=RED,
            label="Negative",
            stroke_width=self.stroke_width,
            font_size=self.font_size,
            buff=self.buff,
        )
        # Distribute the groups uniformly vertically
        assets = Group(anchor_group, positive_group, negative_group)
        assets.arrange(DOWN, buff=1.5)

        return assets

    @override_animation(Create)
    def _create_override(self):
        # TODO make Create animation that is custom
        return FadeIn(self.assets)

    def make_forward_pass_animation(self, layer_args={}, **kwargs):
        """Forward pass for triplet"""
        return AnimationGroup()



"""
    Module for handling syncing two animations one to one.
    The goal here is to zip up two classes and their respective animations,
    and create a joint class with the same animations that runs the animations
    for both classes at the same time. This way we can connect two isomorphic
    views of the same concept and visualize them at the same time. 
"""


class OneToOneSync:
    pass



from manim import *


class LazyAnimation(Animation):
    def __init__(self, animation_function):
        self.animation_function = animation_function
        super.__init__()

    def begin(self):
        update_func_anim = UpdateFromFunc(self.neural_network, create_new_connective)
        self.add

        super.begin()



from manim import *
from manim_ml.neural_network.layers.feed_forward import FeedForwardLayer
from manim_ml.neural_network.layers.parent_layers import ConnectiveLayer
from manim_ml.neural_network.layers.triplet import TripletLayer


class TripletToFeedForward(ConnectiveLayer):
    """TripletLayer to FeedForward layer"""

    input_class = TripletLayer
    output_class = FeedForwardLayer

    def __init__(
        self,
        input_layer,
        output_layer,
        animation_dot_color=RED,
        dot_radius=0.02,
        **kwargs
    ):
        super().__init__(input_layer, output_layer, **kwargs)
        self.animation_dot_color = animation_dot_color
        self.dot_radius = dot_radius

        self.feed_forward_layer = output_layer
        self.triplet_layer = input_layer

    def construct_layer(
        self,
        input_layer: "NeuralNetworkLayer",
        output_layer: "NeuralNetworkLayer",
        **kwargs
    ):
        return super().construct_layer(input_layer, output_layer, **kwargs)

    def make_forward_pass_animation(self, layer_args={}, **kwargs):
        """Makes dots diverge from the given location and move to the feed forward nodes decoder"""
        animations = []
        # Loop through each image
        images = [
            self.triplet_layer.anchor,
            self.triplet_layer.positive,
            self.triplet_layer.negative,
        ]
        for image_mobject in images:
            image_animations = []
            dots = []
            # Move dots from each image to the centers of each of the nodes in the FeedForwardLayer
            image_location = image_mobject.get_center()
            for node in self.feed_forward_layer.node_group:
                new_dot = Dot(
                    image_location,
                    radius=self.dot_radius,
                    color=self.animation_dot_color,
                )
                per_node_succession = Succession(
                    Create(new_dot),
                    new_dot.animate.move_to(node.get_center()),
                )
                image_animations.append(per_node_succession)
                dots.append(new_dot)

            animations.append(AnimationGroup(*image_animations))

        animation_group = AnimationGroup(*animations)

        return animation_group

    @override_animation(Create)
    def _create_override(self):
        return AnimationGroup()



"""
    Animated flow charts. 
"""
from manim import *


class FlowGraph(VGroup):
    """Graph container"""

    pass


class FlowNode(VGroup):
    """Node in the FlowGraph"""

    pass


class DataNode(FlowNode):
    """Node that outputs data"""

    pass



from setuptools import setup, find_packages

setup(
    name="manim_ml",
    version="0.0.13",
    description=(" Machine Learning Animations in python using Manim."),
    packages=find_packages(),
)
